from itertools import chain, repeat
from typing import List, Tuple

import numpy as np
import pandas as pd

from mqboost.base import AlphaLike, ValidationException, XdataLike, YdataLike


def alpha_validate(
    alphas: AlphaLike,
) -> List[float]:
    """
    Validate alphas
    Args:
        alphas (AlphaLike)

    Returns:
        List[float]
    """
    if isinstance(alphas, float):
        alphas = [alphas]

    _len_alphas = len(alphas)
    if _len_alphas == 0:
        raise ValidationException("Input alpha is not valid")

    if _len_alphas >= 2 and any(
        alphas[i] > alphas[i + 1] for i in range(_len_alphas - 1)
    ):
        raise ValidationException("Alpha is not ascending order")

    if _len_alphas != len(set(alphas)):
        raise ValidationException("Duplicated alpha exists")

    return alphas


def prepare_x(
    x: XdataLike,
    alphas: List[float],
) -> pd.DataFrame:
    """
    Return stacked X
    Args:
        x (XdataLike)
        alphas (List[float])

    Returns:
        pd.DataFrame
    """
    if isinstance(x, np.ndarray) or isinstance(x, pd.Series):
        x = pd.DataFrame(x)

    if "_tau" in x.columns:
        raise ValidationException("Column name '_tau' is not allowed.")

    _alpha_repeat_count_list = [list(repeat(alpha, len(x))) for alpha in alphas]
    _alpha_repeat_list = list(chain.from_iterable(_alpha_repeat_count_list))

    _repeated_x = pd.concat([x] * len(alphas), axis=0)
    _repeated_x = _repeated_x.assign(
        _tau=_alpha_repeat_list,
    )
    return _repeated_x


def prepare_train(
    x: XdataLike,
    y: YdataLike,
    alphas: List[float],
) -> Tuple[pd.DataFrame, np.ndarray]:
    """
    Return stacked X, y for training
    Args:
        x (XdataLike)
        y (YdataLike)
        alphas (List[float])

    Returns:
        Tuple[pd.DataFrame, np.ndarray]
    """
    _train_df = prepare_x(x=x, alphas=alphas)
    _repeated_y = np.concatenate(list(repeat(y, len(alphas))))
    return (_train_df, _repeated_y)


def delta_validate(delta: float) -> None:
    _delta_upper_bound: float = 0.1

    if not isinstance(delta, float):
        raise ValidationException("delta is not float type")

    if delta > _delta_upper_bound:
        raise ValidationException("Delta must be smaller than 0.1")



from typing import Any, Callable, Dict, List, Optional

import lightgbm as lgb
import numpy as np
import optuna
import pandas as pd
import xgboost as xgb

from mqboost.base import (
    FUNC_TYPE,
    AlphaLike,
    FittingException,
    ModelName,
    MQStr,
    ObjectiveName,
    TypeName,
    ValidationException,
    XdataLike,
    YdataLike,
)
from mqboost.constraints import set_monotone_constraints
from mqboost.hpo import get_params, train_valid_split
from mqboost.objective import MQObjective
from mqboost.utils import alpha_validate, prepare_train, prepare_x

__all__ = ["MQRegressor"]


class MQRegressor:
    """
    Monotone quantile regressor which preserving monotonicity among quantiles
    Attributes
    ----------
    x: XdataLike
    y: YdataLike
    alphas: AlphaLike
        It must be in ascending order and not contain duplicates.
    objective: str
        Determine objective function. Defaults to "check", another option is "huber".
    model: str
        Determine base model. Defaults to "lightgbm", another option is "xgboost".
    delta (float, optional).
        Only used with "huber" objective.
        Defaults to 0.05 and must be smaller than 0.1.

    Methods
    ----------
    train
    predict
    optimize_params
    """

    def __init__(
        self,
        x: XdataLike,
        y: YdataLike,
        alphas: AlphaLike,
        model: str = ModelName.lightgbm,
        objective: str = ObjectiveName.check,
        delta: float = 0.05,
    ) -> None:
        self._alphas = alpha_validate(alphas)
        self._model = ModelName().get(model)
        self._objective = ObjectiveName().get(objective)
        _funcs = FUNC_TYPE.get(model)
        self._train_dtype: Callable = _funcs.get(TypeName.train_dtype)
        self._predict_dtype: Callable = _funcs.get(TypeName.predict_dtype)
        self._constraints_type: Callable = _funcs.get(TypeName.constraints_type)
        self._MQObj = MQObjective(
            alphas=self._alphas,
            objective=self._objective,
            model=self._model,
            delta=delta,
        )
        self.x_train, self.y_train = prepare_train(x=x, y=y, alphas=self._alphas)
        self.dataset = self._train_dtype(data=self.x_train, label=self.y_train)

    def train(
        self,
        params: Optional[Dict[str, Any]] = None,
        n_trials: int = 20,
    ) -> None:
        """
        Train regressor and return model
        Args:
            params (Optional[Dict[str, Any]])
                Train parameter. Default to None.
                If None, hyperparameter optimization process is executed.
            n_trials (int):
                The number of hyperparameter tuning. Default to 20.
        """
        if params is None:
            params = self.optimize_params(n_trials=n_trials)

        self._params = set_monotone_constraints(
            params=params,
            x_train=self.x_train,
            constraints_fucs=self._constraints_type,
        )

        if self.__is_lgb:
            self._params.update({MQStr.obj: self._MQObj.fobj})
            self.model = lgb.train(
                train_set=self.dataset,
                params=self._params,
                feval=self._MQObj.feval,
                valid_sets=[self.dataset],
            )
        elif self.__is_xgb:
            self.model = xgb.train(
                dtrain=self.dataset,
                verbose_eval=False,
                params=self._params,
                obj=self._MQObj.fobj,
                custom_metric=self._MQObj.feval,
                evals=[(self.dataset, "train")],
            )
        else:
            raise FittingException("model name is invalid")
        self._fitted = True

    def predict(
        self,
        x: XdataLike,
        alphas: Optional[AlphaLike] = None,
    ) -> np.ndarray:
        """
        Return predicted quantiles
        Args:
            x (XdataLike)
            alphas (Optional[AlphaLike], optional). Defaults to None.

        Returns:
            np.ndarray
        """
        self.__is_fitted
        if alphas is None:
            alphas = self._alphas
        else:
            alphas = alpha_validate(alphas=alphas)
        _x = prepare_x(x=x, alphas=alphas)
        _x = self._predict_dtype(_x)
        _pred = self.model.predict(data=_x)
        _pred = _pred.reshape(len(alphas), len(x))
        return _pred

    def optimize_params(
        self,
        n_trials: int,
        get_params_func: Callable = get_params,
        valid_dict: Optional[Dict[str, XdataLike]] = None,
    ) -> Dict[str, Any]:
        """
        Optimize hyperparameter
        Args:
            n_trials (int): The number of trials for the hyperparameter optimization.
            get_params_func (Callable, optional):
                A function to get the parameters for the model.
                For example,
                    def get_params(trial: Trial, model: ModelName):
                        return {
                            "learning_rate": trial.suggest_float("learning_rate", 1e-2, 1.0, log=True),
                            "max_depth": trial.suggest_int("max_depth", 1, 10),
                            "lambda_l1": trial.suggest_float("lambda_l1", 1e-8, 10.0, log=True),
                            "lambda_l2": trial.suggest_float("lambda_l2", 1e-8, 10.0, log=True),
                            "num_leaves": trial.suggest_int("num_leaves", 2, 256),
                            "feature_fraction": trial.suggest_float("feature_fraction", 0.4, 1.0),
                            "bagging_fraction": trial.suggest_float("bagging_fraction", 0.4, 1.0),
                            "bagging_freq": trial.suggest_int("bagging_freq", 1, 7),
                        }
            valid_dict (Dict):
                Manually selected validation set. Keys must contain "data" and "label".
                For example,
                    valid_dict = {
                        "data": ...,
                        "label": ...,
                    }
        Returns:
            Dict[str, Any]: best params
        """
        if valid_dict is None:
            x_train, x_valid, y_train, y_valid = train_valid_split(
                x_train=self.x_train, y_train=self.y_train
            )
        else:
            _valid_dict_keys: List[str] = ["data", "label"]
            if any([_ not in valid_dict.keys() for _ in _valid_dict_keys]):
                raise ValidationException(
                    "Key of valid_dict must contains 'data' and 'label'"
                )
            x_train = self.x_train
            y_train = self.y_train
            _x_valid = valid_dict.get("data")
            _y_valid = valid_dict.get("label")
            x_valid, y_valid = prepare_train(
                x=_x_valid, y=_y_valid, alphas=self._alphas
            )

        def _study_func(trial: optuna.Trial) -> float:
            return self.__optuna_objective(
                trial=trial,
                x_train=x_train,
                x_valid=x_valid,
                y_train=y_train,
                y_valid=y_valid,
                constraints_func=self._constraints_type,
                get_params_func=get_params_func,
            )

        self._study = optuna.create_study(
            study_name=f"MQBoost_{self._model}",
            direction="minimize",
            load_if_exists=True,
        )
        self._study.optimize(_study_func, n_trials=n_trials)
        return self._study.best_params

    def __optuna_objective(
        self,
        trial: optuna.Trial,
        x_train: pd.DataFrame,
        x_valid: pd.DataFrame,
        y_train: np.ndarray,
        y_valid: np.ndarray,
        constraints_func: Callable,
        get_params_func: Callable,
    ) -> float:
        """
        objective function for optuna
        Args:
            trial (optuna.Trial)
            x_train (pd.DataFrame)
            x_valid (pd.DataFrame)
            y_train (np.ndarray)
            y_valid (np.ndarray)
            constraints_func (Callable)
            get_params_func (Callable)
        Returns:
            float
        """
        params = get_params_func(trial, self._model)
        params = set_monotone_constraints(
            params, x_train=x_train, constraints_fucs=constraints_func
        )
        dtrain = self._train_dtype(data=x_train, label=y_train)
        dvalid = self._train_dtype(data=x_valid, label=y_valid)
        if self.__is_lgb:
            model_params = dict(
                params=params,
                train_set=dtrain,
                valid_sets=dvalid,
            )
            _gbm = lgb.train(**model_params)
            _preds = _gbm.predict(data=x_valid, num_iteration=_gbm.best_iteration)
            _, loss, _ = self._MQObj.feval(y_pred=_preds, dtrain=dvalid)
        elif self.__is_xgb:
            model_params = dict(
                params=params,
                dtrain=dtrain,
                evals=[(dvalid, MQStr.valid)],
            )
            _gbm = xgb.train(**model_params)
            _preds = _gbm.predict(data=self._predict_dtype(x_valid))
            _, loss = self._MQObj.feval(y_pred=_preds, dtrain=dvalid)
        else:
            raise FittingException("model name is invalid")
        return loss

    @property
    def MQObj(self) -> MQObjective:
        return self._MQObj

    @property
    def study(self) -> optuna.Study:
        return getattr(self, "_study", None)

    @property
    def __is_lgb(self) -> bool:
        return self._model == ModelName.lightgbm

    @property
    def __is_xgb(self) -> bool:
        return self._model == ModelName.xgboost

    @property
    def __is_fitted(self) -> None:
        if not getattr(self, "_fitted", False):
            raise FittingException("train must be executed before predict")



from typing import Any, Callable, Dict

from mqboost.base import MQStr, ValidationException, XdataLike


def set_monotone_constraints(
    params: Dict[str, Any], x_train: XdataLike, constraints_fucs: Callable
) -> Dict[str, Any]:
    """
    Set monotone constraints in params
    Args:
        params (Dict[str, Any])
        x_train (XdataLike)
        constraints_fucs (Callable)

    Raises:
        ValidationException: when "objective" is in params.keys()

    Returns:
        Dict[str, Any]
    """
    if MQStr.obj in params:
        raise ValidationException(
            "The parameter named 'objective' must be excluded in params"
        )
    _params = params.copy()
    if MQStr.mono in _params:
        _monotone_constraints = list(_params[MQStr.mono])
        _monotone_constraints.append(1)
        _params[MQStr.mono] = constraints_fucs(_monotone_constraints)
    else:
        _params.update(
            {
                MQStr.mono: constraints_fucs(
                    [1 if "_tau" == col else 0 for col in x_train.columns]
                )
            }
        )
    return _params



from collections.abc import Callable
from dataclasses import dataclass
from typing import Dict, List, Union

import lightgbm as lgb
import numpy as np
import pandas as pd
import xgboost as xgb


# Name
class BaseName:
    def get(self, name):
        if name in self.__annotations__:
            return getattr(self, name)
        else:
            available_attributes = list(self.__annotations__.keys())
            raise AttributeError(
                f"'{self.__class__.__name__}' object has no attribute '{name}'. "
                f"Available attributes are: {', '.join(available_attributes)}"
            )


@dataclass
class ModelName(BaseName):
    lightgbm: str = "lightgbm"
    xgboost: str = "xgboost"


@dataclass
class ObjectiveName(BaseName):
    check: str = "check"
    huber: str = "huber"


@dataclass
class TypeName(BaseName):
    train_dtype: str = "train_dtype"
    predict_dtype: str = "predict_dtype"
    constraints_type: str = "constraints_type"


@dataclass
class MQStr(BaseName):
    mono: str = "monotone_constraints"
    obj: str = "objective"
    valid: str = "valid"


# Functions
FUNC_TYPE: Dict[str, Dict[str, Callable]] = {
    ModelName.lightgbm: {
        TypeName.train_dtype: lgb.Dataset,
        TypeName.predict_dtype: lambda x: x,
        TypeName.constraints_type: list,
    },
    ModelName.xgboost: {
        TypeName.train_dtype: xgb.DMatrix,
        TypeName.predict_dtype: xgb.DMatrix,
        TypeName.constraints_type: tuple,
    },
}


# Type
XdataLike = Union[pd.DataFrame, pd.Series, np.ndarray]
YdataLike = Union[pd.Series, np.ndarray]
AlphaLike = Union[List[float], float]
ModelLike = Union[lgb.basic.Booster, xgb.Booster]
DtrainLike = Union[lgb.basic.Dataset, xgb.DMatrix]


# Exception
class FittingException(Exception):
    pass


class ValidationException(Exception):
    pass



from typing import Tuple

import numpy as np
import pandas as pd
from optuna import Trial
from sklearn.model_selection import train_test_split

from mqboost.base import FittingException, ModelName


def get_params(trial: Trial, model: ModelName):
    if model == ModelName.lightgbm:
        params = {
            "verbose": -1,
            "learning_rate": trial.suggest_float("learning_rate", 1e-2, 1.0, log=True),
            "max_depth": trial.suggest_int("max_depth", 1, 10),
            "lambda_l1": trial.suggest_float("lambda_l1", 1e-8, 10.0, log=True),
            "lambda_l2": trial.suggest_float("lambda_l2", 1e-8, 10.0, log=True),
            "num_leaves": trial.suggest_int("num_leaves", 2, 256),
            "feature_fraction": trial.suggest_float("feature_fraction", 0.4, 1.0),
            "bagging_fraction": trial.suggest_float("bagging_fraction", 0.4, 1.0),
            "bagging_freq": trial.suggest_int("bagging_freq", 1, 7),
        }
    elif model == ModelName.xgboost:
        params = {
            "learning_rate": trial.suggest_float("learning_rate", 1e-2, 1.0, log=True),
            "max_depth": trial.suggest_int("max_depth", 1, 10),
            "reg_lambda": trial.suggest_float("reg_lambda", 1e-8, 100.0, log=True),
            "reg_alpha": trial.suggest_float("reg_alpha", 1e-8, 100.0, log=True),
            "subsample": trial.suggest_float("subsample", 0.1, 1.0),
            "colsample_bytree": trial.suggest_float("colsample_bytree", 0.1, 1.0),
        }
    else:
        raise FittingException("model name is invalid")
    return params


def train_valid_split(
    x_train: pd.DataFrame, y_train: np.ndarray
) -> Tuple[pd.DataFrame, pd.DataFrame, np.ndarray, np.ndarray]:
    x_train, x_valid, y_train, y_valid = train_test_split(
        x_train, y_train, test_size=0.2, random_state=42, stratify=x_train["_tau"]
    )
    return x_train, x_valid, y_train, y_valid



from collections.abc import Callable
from functools import partial
from typing import Any, List, Tuple

import numpy as np

from mqboost.base import DtrainLike, ModelName, ObjectiveName
from mqboost.utils import delta_validate

CHECK_LOSS: str = "check_loss"


def _grad_rho(u: np.ndarray, alpha: float) -> np.ndarray:
    return (u < 0).astype(int) - alpha


def _rho(u: np.ndarray, alpha: float) -> np.ndarray:
    return -u * _grad_rho(u=u, alpha=alpha)


def _error_delta_compare(u: np.ndarray, delta: float):
    _abs_error = np.abs(u)
    return (_abs_error <= delta).astype(int), (_abs_error > delta).astype(int)


def _grad_huber(u: np.ndarray, alpha: float, delta: float) -> np.ndarray:
    _smaller_delta, _bigger_delta = _error_delta_compare(u=u, delta=delta)
    _g = _grad_rho(u=u, alpha=alpha)
    _r = _rho(u=u, alpha=alpha)
    return _r * _smaller_delta + _g * _bigger_delta


def _train_pred_reshape(
    y_pred: np.ndarray,
    dtrain: DtrainLike,
    len_alpha: int,
) -> Tuple[np.ndarray, np.ndarray]:
    _y_train: np.ndarray = dtrain.get_label()
    return _y_train.reshape(len_alpha, -1), y_pred.reshape(len_alpha, -1)


def _compute_grads_hess(
    y_pred: np.ndarray,
    dtrain: DtrainLike,
    alphas: List[float],
    grad_fn: Callable[[np.ndarray, float, Any], np.ndarray],
    **kwargs: Any,
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Compute gradients for given loss function
    Args:
        y_train (np.ndarray)
        y_pred (np.ndarray)
        alphas (List[float])
        grad_fn (Callable[[np.ndarray, float, Any], np.ndarray])
        **kwargs (Any): Additional arguments for grad_fn

    Returns:
        np.ndarray
    """
    _len_alpha = len(alphas)
    _y_train, _y_pred = _train_pred_reshape(
        y_pred=y_pred, dtrain=dtrain, len_alpha=_len_alpha
    )
    grads = []
    for alpha_inx in range(len(alphas)):
        _err_for_alpha = _y_train[alpha_inx] - _y_pred[alpha_inx]
        _grad = grad_fn(u=_err_for_alpha, alpha=alphas[alpha_inx], **kwargs)
        grads.append(_grad)
    return np.concatenate(grads), np.ones_like(y_pred)


check_loss_grad_hess: Callable = partial(_compute_grads_hess, grad_fn=_grad_rho)
huber_loss_grad_hess: Callable = partial(_compute_grads_hess, grad_fn=_grad_huber)


def eval_loss(
    y_pred: np.ndarray,
    dtrain: DtrainLike,
    alphas: List[float],
) -> Tuple[str, float]:
    """
    eval funcs
    Args:
        y_pred (np.ndarray)
        d_train (DtrainLike)
        alphas (List[float])
    Returns:
        Tuple[str, float]
    """
    _len_alpha = len(alphas)
    _y_train, _y_pred = _train_pred_reshape(
        y_pred=y_pred, dtrain=dtrain, len_alpha=_len_alpha
    )
    loss: float = 0.0
    for alpha_inx in range(_len_alpha):
        _err_for_alpha = _y_train[alpha_inx] - _y_pred[alpha_inx]
        _loss = _rho(u=_err_for_alpha, alpha=alphas[alpha_inx])
        loss = loss + np.mean(_loss)

    loss = loss / _len_alpha
    return CHECK_LOSS, loss


def _lgb_eval_loss(
    y_pred: np.ndarray,
    dtrain: DtrainLike,
    alphas: List[float],
) -> Tuple[str, np.ndarray, bool]:
    loss_str, loss = eval_loss(y_pred=y_pred, dtrain=dtrain, alphas=alphas)
    return loss_str, loss, False


class MQObjective:
    """
    Monotone quantile objective and evaluation function
    Attributes
    ----------
    alphas (List[float])
    objective (ObjectiveName)
    model (ModelName)
    delta (float)

    Property
    ----
    fobj
    feval
    eval_name
    """

    def __init__(
        self,
        alphas: List[float],
        objective: ObjectiveName,
        model: ModelName,
        delta: float,
    ) -> None:
        if objective == ObjectiveName.huber:
            delta_validate(delta=delta)
            self._fobj = partial(huber_loss_grad_hess, alphas=alphas, delta=delta)
        elif objective == ObjectiveName.check:
            self._fobj = partial(check_loss_grad_hess, alphas=alphas)

        self._eval_name = CHECK_LOSS
        if model == ModelName.lightgbm:
            self._feval = partial(_lgb_eval_loss, alphas=alphas)
        elif model == ModelName.xgboost:
            self._feval = partial(eval_loss, alphas=alphas)

    @property
    def fobj(self) -> Callable:
        return self._fobj

    @property
    def feval(self) -> Callable:
        return self._feval

    @property
    def eval_name(self) -> str:
        return self._eval_name
